from yafw.project_management import FrealignJob, FrealignProject, FrealignParameters
from pathlib import Path
import re
from pydantic import BaseModel
import os

#https://stackoverflow.com/questions/2301789/how-to-read-a-file-in-reverse-order
def reverse_readline(filename, buf_size=8192):
    """A generator that returns the lines of a file in reverse order"""
    with open(filename, 'rb') as fh:
        segment = None
        offset = 0
        fh.seek(0, os.SEEK_END)
        file_size = remaining_size = fh.tell()
        while remaining_size > 0:
            offset = min(file_size, offset + buf_size)
            fh.seek(file_size - offset)
            buffer = fh.read(min(remaining_size, buf_size))
            # remove file's last "\n" if it exists, only for the first buffer
            if remaining_size == file_size and buffer[-1] == ord('\n'):
                buffer = buffer[:-1]
            remaining_size -= buf_size
            lines = buffer.split('\n'.encode())
            # append last chunk's segment to this chunk's last line
            if segment is not None:
                lines[-1] += segment
            segment = lines[0]
            lines = lines[1:]
            # yield lines in this chunk except the segment
            for line in reversed(lines):
                # only decode on a parsed line, to avoid utf-8 decode error
                yield line.decode()
        # Don't yield None if the file was empty
        if segment is not None:
            yield segment.decode()



class FSCCurve(BaseModel):
    ring_radius: list[float] = []
    resolution: list[float]= []
    fsc_values: list[float]= []
    part_fsc_values: list[float] = []

class FrealignResults(BaseModel):
    round: int
    class_n: int
    avg_occ: float = None
    avg_score: float = None
    FSC: FSCCurve = FSCCurve()



def continue_job(project: FrealignProject, job: FrealignJob, nrounds: int = 1):
    import shutil
    current_params = FrealignParameters.open(job.path / "mparameters")
    shutil.copy(job.path / "mparameters", job.path / f"mparameters_{current_params.start_process}_{current_params.end_process}")
    current_params.start_process = current_params.end_process + 1
    current_params.end_process = current_params.start_process + nrounds - 1 
    current_params.render(job.path / "mparameters")

#def combine_classes(project: FrealignProject, 
#                    name: str, 
#                    job: FrealignJob,
#                    class_numbers: list[int],
#                    iteration: int = -1):
    

def parse_job(project: FrealignProject, job: FrealignJob):
    from rich.progress import track
    par_files = list(job.path.glob(f"*.par"))
    results = []
    for par_file in track(par_files):
        if project is None:
            regexp = re.compile(r".+_(\d+)_r(\d+).par")
        else :
            regexp = re.compile(f"{project.name}_(\d+)_r(\d+).par")        
        match = re.match(regexp, par_file.name)

        if match:
            round_n = int(match.group(1))
            class_n = int(match.group(2))
            result = FrealignResults(round=round_n, class_n=class_n)
            results.append(result)
            # Read the par file and extract the occupancy values
            read_state = "NONE"
            for line in reverse_readline(par_file):
                if not line.startswith("C"):
                    break
                if line.startswith("C  Average"):
                    read_state = "FSC"
                    continue
                if line.startswith("C  NO"):
                    read_state = "NONE"
                    continue
                if line.startswith("C  Total"):
                    splitline = line.split()
                    result.avg_score = float(splitline[-2])
                    result.avg_occ = float(splitline[-1])
                    continue
                if read_state == "FSC":
                    splitline = line.split()
                    result.FSC.ring_radius.append(float(splitline[3]))
                    result.FSC.resolution.append(float(splitline[2]))
                    result.FSC.fsc_values.append(float(splitline[5]))
                    result.FSC.part_fsc_values.append(float(splitline[6]))
                    continue
    results.sort(key=lambda x: (x.round, x.class_n))
    per_class_results = []
    for class_n in range(1, results[-1].class_n+1):
        round_results = [result for result in results if result.class_n == class_n]
        per_class_results.append(round_results)
    return per_class_results


from pydantic import BaseModel
from pathlib import Path
import typer
from typing import Optional
from enum import Enum

class FrealignJobStatus(Enum):
    PREPARED = "prepared"
    RUNNING = "running"
    FINISHED = "finished"
    FAILED = "failed"

class FrealignBinnedStack(BaseModel):
    filename: Path
    pixel_size_A: float

class FrealignJob(BaseModel):
    id: int
    path: Path
    status: FrealignJobStatus = FrealignJobStatus.PREPARED

class FrealignProject(BaseModel):
    name: str
    path: Path
    imported_starfile: Path
    imported_mrcfile: Path
    original_pixelsize_A: float
    detector_pixelsize_A: float = 50000.0
    stacks: list[FrealignBinnedStack] = []
    jobs: list[FrealignJob] = []


    def save(self):
        with open(self.path / f"{self.name}.json", "w") as f:
            f.write(self.model_dump_json())
    
    @classmethod
    def open(cls, filename):
        with open(filename, "r") as f:
            data = f.read()
        return cls.model_validate_json(data)

class Global(BaseModel):
    project: Optional[FrealignProject]
    job: Optional[FrealignJob]
    

class Context(typer.Context):
    obj: Optional[Global] = None

class FrealignParameters(BaseModel):
    cluster_type: str = "none"
    nprocessor_ref: int = 10
    nprocessor_rec: int = 10
    mem_per_cpu: int = 2048
    MODE: int = 1
    start_process: int = 2
    end_process: int = 10
    res_high_refinement: float = 16.0
    res_high_class: float = 16.0
    nclasses: int = 10
    data_input: str = "stack"
    raw_images: str = "stack.mrc"
    image_contrast: str = "N"
    outer_radius: float = 200.0
    inner_radius: float = 0.0
    mol_mass: float = 1000
    Symmetry: str = "C1"
    pix_size: float = 8.0
    dstep: float = 37.7
    Aberration: float = 2.7
    Voltage: float = 300.0
    Amp_contrast: float = 0.07
    mask_file: str = ""
    mask_edge: int = 5
    mask_outside_weight: float = 0.0
    mask_filt_res: float = 0.0
    mask_filt_edge: int = 5
    focus_mask: str = ""
    DANG: float = 0.0
    parameter_mask: str = "0 0 0 0 0"
    refineangleinc: int = 4
    refineshiftinc: int = 4
    res_reconstruction: float = 0.0
    res_low_refinement: float = 50.0
    thresh_reconst: float = 0.0
    nbootstrap: int = 1000
    FDEF: str = 'F'
    FMATCH: str = 'F'
    FBOOST: str = 'T'
    mp_cpus: int = 1
    restart_after_crash: str = 'F'
    delete_scratch: str = 'T'
    qsub_string_ref: str = ""
    qsub_string_rec: str = ""
    first_particle: str = ""
    last_particle: str = ""
    frealign_bin_dir: str = ""
    scratch_dir: str = ""
    ITMAX: int = 200
    XSTD: float = 0.0
    PBC: float = 2.0
    thresh_refine: float = 50.0
    RBfactor: float = 0.0
    FMAG: str = 'F'
    FASTIG: str = 'F'
    FPART: str = 'F'
    FFILT: str = 'T'
    FBEAUT: str = 'F'
    beam_tilt_x: float = 0.0
    beam_tilt_y: float = 0.0
    BSC: float = 2.0
    percentage: float = 1.0
    normalize_images: str = 'F'
    crop_images: str = 'F'
    adjust_scores: str = 'T'
    search_mask_radius: float = 0.0
    search_range_x: float = 0.0
    search_range_y: float = 0.0
    boost_limit: float = 0.0
    P3Dblurring: str = 'F'
    even_odd_fsc: str = 'T'

    def render(self, filename):
        with open(filename, "w") as f:
            for k, v in self.model_dump().items():
                if k in ["parameter_mask", "focus_mask"]:
                    f.write(f'{k}\t\t"{v}"\n')
                else:
                    f.write(f'{k}\t\t{v}\n')
    
    @classmethod
    def open(cls, filename):
        import shlex
        with open(filename, "r") as f:
            data = f.readlines()
        out = cls()
        for line in data:
            k, *v = shlex.split(line)
            if len(v) > 0:
                setattr(out, k, v[0])
        out = cls(**out.model_dump())
        return out
            


import typer
from pathlib import Path
from yafw.project_management import Context, Global
app = typer.Typer()

@app.command()
def create_project(
    starfilename: Path = typer.Argument(..., help="Path to the starfile"),
    mrcfilename: Path = typer.Argument(..., help="Path to the mrcfile"),
    name: str = typer.Argument("stack", help="Name of the project"),
    original_pixelsize: float = typer.Argument(..., help="Physical pixelsize that the data was collected at in Angstroms"),
    detector_pixelsize: float = typer.Option(50000.0, help="Physical pixelsize of the detector in Angstroms"),
):
    import starfile
    from shutil import copy
    from yafw.project_management import FrealignProject, FrealignBinnedStack

    project_dir = Path.cwd() / name
    project_dir.mkdir()
    copy(mrcfilename, project_dir/f"{name}.mrc")

    particle_data = starfile.read(starfilename)

    with open(project_dir / f"{name}.par", "w") as f:
        for i, particle in enumerate(particle_data.itertuples()):
            shift_x = 0.0
            shift_y = 0.0
            if hasattr(particle, "cisTEMShiftX"):
                shift_x = particle.cisTEMShiftX
                shift_y = particle.cisTEMShiftY
            f.write("%7d%8.2f%8.2f%8.2f%10.2f%10.2f%8d%6d%9.1f%9.1f%8.2f%8.2f%10d%11.4f%8.2f%8.2f\n" % (
                particle.cisTEMPositionInStack,
                particle.cisTEMAnglePsi,
                particle.cisTEMAngleTheta,
                particle.cisTEMAnglePhi,
                shift_x,
                shift_y,
                detector_pixelsize / original_pixelsize,
                1,
                particle.cisTEMDefocus1,
                particle.cisTEMDefocus2,
                particle.cisTEMDefocusAngle,
                100.0,
                0,
                0.5,
                0.0,
                0.0
            ))
    #"%7d%8.2f%8.2f%8.2f%10.2f%10.2f%8d%6d%9.1f%9.1f%8.2f%8.2f%10d%11.4f%8.2f%8.2f\n"
    # do something with the starfile and mrcfile
    project = FrealignProject(
        name=name,
        path=project_dir,
        imported_starfile=starfilename,
        imported_mrcfile=mrcfilename,
        original_pixelsize_A=original_pixelsize,
        detector_pixelsize_A=detector_pixelsize
    )

    project.stacks.append(FrealignBinnedStack(
        filename=project_dir / f"{name}.mrc",
        pixel_size_A=particle_data.cisTEMPixelSize[0]
    ))
    # Write the project to disk as json
    project.save()
    typer.echo(f"Created parfile {name}.par and stack {name}.mrc in {project_dir}")

@app.command()
def create_binned_stack(
    ctx: Context,
    bin: int = typer.Argument(..., help="Factor to bin by"),
):
    if ctx.obj is None:
        typer.echo("Please run in a FREALIGN project folder. Exiting.")
        raise typer.Exit(code=1)
    
    from pycistem.programs.resample import ResampleParameters, run
    import mrcfile
    from yafw.project_management import FrealignBinnedStack

    with mrcfile.open(ctx.obj.project.stacks[0].filename, mode="r") as f:
        orig_size = f.header.nx
    bin_to = round(orig_size / bin)
    output_filename = ctx.obj.project.path / f"{ctx.obj.project.name}B{bin}.mrc"
    if output_filename.exists():
        typer.echo(f"{output_filename} already exists. Exiting.")
        raise typer.Exit(code=1)
    par = ResampleParameters(
        input_filename=str(ctx.obj.project.stacks[0].filename),
        output_filename=str(output_filename),
        new_x_size=bin_to,
        new_y_size=bin_to,
    )
    print(f"Creating binned stack {output_filename}")
    run(par)
    ctx.obj.project.stacks.append(FrealignBinnedStack(
        filename=output_filename,
        pixel_size_A=ctx.obj.project.stacks[0].pixel_size_A * (orig_size / bin_to)
    ))
    ctx.obj.project.save()

@app.command()
def create_job(
    ctx: Context,
    stack: Path = typer.Argument(..., help="The stack to use"),
    starting_parameters: str = typer.Argument("default", help="Path to the starting parameters file"),
    parameter_mask: str = typer.Option("0 0 0 0 0", help="Path to the mask file"),
    nclasses: int = typer.Option(10, help="Number of classes to refine"),
    nprocessor_ref: int = typer.Option(10, help="Number of processors to use for refinement"),
    nprocessor_rec: int = typer.Option(10, help="Number of processors to use for reconstruction"),
    nrounds: int = typer.Option(10, help="Number of rounds to run"),
    outer_radius: float = typer.Option(200.0, help="Outer radius of the mask"),
    mol_mass: float = typer.Option(3000.0, help="Molecular mass of the particle"),
):
    if ctx.obj is None:
        typer.echo("Please run in a FREALIGN project folder. Exiting.")
        raise typer.Exit(code=1)
    from yafw.project_management import FrealignJob, FrealignParameters
    from shutil import copy
    job_id = max([job.id for job in ctx.obj.project.jobs], default=0) + 1
    job = FrealignJob(
        id=job_id,
        path=ctx.obj.project.path / f"job{job_id:04d}"
    )
    job.path.mkdir()
    if starting_parameters == "default":
        copy(ctx.obj.project.path / f"{ctx.obj.project.name}.par", job.path / f"{ctx.obj.project.name}_1_r1.par")
    else:
        copy(starting_parameters, job.path / f"{ctx.obj.project.name}_1_r1.par")
    stack_info = list(filter(lambda x: x.filename.absolute() == stack.absolute(),ctx.obj.project.stacks))[0]
    
    parameters = FrealignParameters(
        parameter_mask=parameter_mask,
        nclasses=nclasses,
        nprocessor_ref=nprocessor_ref,
        nprocessor_rec=nprocessor_rec,
        start_process=2,
        end_process=2 + nrounds - 1,
        data_input=ctx.obj.project.name,
        raw_images=str(stack.absolute()),
        outer_radius=outer_radius,
        mol_mass=mol_mass,
        pix_size = stack_info.pixel_size_A,
        dstep=(ctx.obj.project.detector_pixelsize_A * (stack_info.pixel_size_A / ctx.obj.project.original_pixelsize_A))/10000,
    )
    parameters.render(job.path / "mparameters")
    ctx.obj.project.jobs.append(job)
    ctx.obj.project.save()

@app.command()
def fsc_results(ctx: Context):
    from yafw.frealign_jobs import parse_job
    if ctx.obj is None or ctx.obj.job is None:
        typer.echo("Please run in a FREALIGN job folder. Exiting.")
        raise typer.Exit(code=1)
    job_data = parse_job(ctx.obj.project, ctx.obj.job)
    #print(job_data[0][-1])
    import matplotlib.pyplot as plt
    import matplotlib.ticker as tick
    import numpy as np
    for i, class_list in enumerate(job_data):
        last_round = class_list[-1]
        resolution = last_round.FSC.resolution
        FSC_values = last_round.FSC.fsc_values
        data_tuples = list(zip(resolution, FSC_values))
        value_found = False

        for resolution, fsc_value in data_tuples:
            if fsc_value <= 0.143:
                print(f"Resolution: {resolution}, FSC Value: {fsc_value}, Class {i+1}")
                value_found = True

        if not value_found:
            print(f"No value <= 0.143 in the Class {i+1}")

    def on_legend_click(event):
        legend = event.artist
        index = event.ind[0]
        line = legend.get_lines()[index]
        line.set_visible(not line.get_visible())
        plt.draw()
   
    fig = plt.figure()
    ax = fig.add_subplot(111)
    for i, class_list in enumerate(job_data):
        last_round = class_list[-1]
        number_values = len(last_round.FSC.resolution)
        ax.plot(range(number_values),last_round.FSC.fsc_values,label=f"Class {i+1}")
        
        plt.xlim(number_values-1,0)
        plt.ylim(-0.01,1.05)
        def x_fmt(x, y):
            if x>number_values-1:
                return ""
            return last_round.FSC.resolution[int(x)]
    ax.xaxis.set_major_formatter(tick.FuncFormatter(x_fmt))
    plt.axhline(0.143, color="black", ls=":")
    plt.xlabel('Resolution (Å)')
    plt.ylabel('FSC')
    legend = plt.legend()
    #plt.legend()
    plt.gcf().canvas.mpl_connect('pick_event', on_legend_click)
    plt.show()

    fig2 = plt.figure()
    ax = fig2.add_subplot(111)
    for i, class_list in enumerate(job_data):
        last_round = class_list[-1]
        number_values = len(last_round.FSC.resolution)
        ax.plot(range(number_values),last_round.FSC.part_fsc_values,label=f"Class {i+1}")
        
        plt.xlim(number_values-1,0)
        plt.ylim(-0.01,1.05)
        def x_fmt(x, y):
            if x>number_values-1:
                return ""
            return last_round.FSC.resolution[int(x)]
    ax.xaxis.set_major_formatter(tick.FuncFormatter(x_fmt))
    plt.axhline(0.143, color="black", ls=":")
    plt.xlabel('Resolution (Å)')
    plt.ylabel('Part_FSC')
    plt.legend()
    plt.show()  

@app.command()
def job_status(ctx: Context):
    from yafw.frealign_jobs import parse_job
    if ctx.obj is None or ctx.obj.job is None:
        typer.echo("Please run in a FREALIGN job folder. Exiting.")
        raise typer.Exit(code=1)
    job_data = parse_job(ctx.obj.project, ctx.obj.job)

    import matplotlib.pyplot as plt
    for class_n in range(job_data[-1][-1].class_n):
        data = job_data[class_n]
        plt.plot([d.round for d in data], [d.avg_occ for d in data], label=f"Class {class_n+1}")
    plt.legend()
    plt.show()

@app.command()
def continue_job(ctx: Context, nrounds: int = typer.Argument(1, help="Number of rounds to continue")):
    from yafw.frealign_jobs import continue_job
    if ctx.obj is None or ctx.obj.job is None:
        typer.echo("Please run in a FREALIGN job folder. Exiting.")
        raise typer.Exit(code=1)
    continue_job(ctx.obj.project, ctx.obj.job, nrounds=nrounds)

@app.command()
def combine_classes(ctx: Context,
                    name: str = typer.Argument(..., help="Name of the new particle set"),
                    job_id: int = typer.Argument(..., help="Job ID to combine classes from"),
                    class_numbers: list[str] = typer.Argument(..., help="Class numbers to combine"),
                    iteration: int = typer.Option(-1, help="Iteration number to combine classes from, -1 is the latest")):
    
    from yafw.frealign_jobs import combine_classes
    if ctx.obj is None:
        typer.echo("Please run in a FREALIGN project folder. Exiting.")
        raise typer.Exit(code=1)
    job = list(filter(lambda x: x.id == job_id, ctx.obj.project.jobs))[0]
    combine_classes(project=ctx.obj.project, name=name, job=job, class_numbers=class_numbers, iteration=iteration)
    

@app.callback()
def main(
    ctx: Context,
):
    from yafw.project_management import FrealignProject, FrealignJob
    if (Path.cwd() / f"{Path.cwd().name}.json").exists():
        ctx.obj = Global(project=FrealignProject.open(f"{Path.cwd() / Path.cwd().name}.json"),job=None)
    elif (Path.cwd().parent / f"{Path.cwd().parent.name}.json").exists():
        project = FrealignProject.open(f"{Path.cwd().parent / Path.cwd().parent.name}.json")
        job = list(filter(lambda x: x.path.absolute() == Path.cwd().absolute(),project.jobs))
        if len(job) == 1:
            job = job[0]
        else:
            job = FrealignJob(
                id=0,
                path=Path.cwd()
            )
        ctx.obj = Global(project=project,
                         job=job)
    else:
        project = None
        job = FrealignJob(
                id=0,
                path=Path.cwd()
            )
        ctx.obj = Global(project=project,
                         job=job)


if __name__ == "__main__":
    app()